# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


from .adapter import to_cqlib
from .api_client import ApiClient
from .gates import X2PGate, X2MGate, Y2PGate, Y2MGate, XY2MGate, XY2PGate
from .job import TianYanJob
from .sampler import TianYanSampler
from .tianyan_provider import TianYanProvider
from .tianyan_backend import TianYanBackend

__all__ = [
    "to_cqlib",
    "ApiClient",

    'X2PGate',
    'X2MGate',
    'Y2PGate',
    'Y2MGate',
    'XY2PGate',
    'XY2MGate',

    "TianYanJob",
    "TianYanSampler",
    "TianYanProvider",
    "TianYanBackend",
]



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
TianYan Backend module for managing quantum computing backends.

This module provides classes to handle backend configurations, properties, and operations
for both quantum computers and simulators on the TianYan platform.
"""

import json
import warnings
from collections import namedtuple
from datetime import datetime
from enum import Enum, IntEnum

from qiskit.circuit import QuantumCircuit, Parameter, Measure, Barrier
from qiskit.circuit.library import standard_gates
from qiskit.circuit.library.standard_gates import CZGate, RZGate, HGate, \
    GlobalPhaseGate, CXGate
from qiskit.providers import BackendV2 as Backend, Options, JobV1, QubitProperties
from qiskit.transpiler import Target, InstructionProperties

from .api_client import ApiClient
from .adapter import to_cqlib
from .gates import X2PGate, X2MGate, Y2MGate, Y2PGate, XY2MGate, XY2PGate
from .job import TianYanJob


class BackendType(IntEnum):
    """Enum representing the type of backend (quantum computer or simulator)."""
    quantum_computer = 0
    simulator = 1
    offline_simulator = 2


class BackendStatus(IntEnum):
    """Enum representing the status of the backend."""
    running = 0
    calibrating = 1
    under_maintenance = 2
    offline = 3


class CqlibAdapterError(Exception):
    """Base class for exceptions in this module."""
    pass


GateConfig = namedtuple('GateConfig', ['name', 'params', 'coupling_map'])

gate_parameters = {
    'rx': 1,
    'ry': 1,
    'rz': 1,
    'rxy': 2,
    'xy2p': 1,
    'xy2m': 1,
}


class BackendConfiguration:
    """Class representing the configuration of a backend."""

    def __init__(
            self,
            backend_id: str,
            backend_name: str,
            n_qubits: int,
            basis_gates: list,
            gates: list,
            local: bool,
            simulator: bool,
            conditional: bool,
            coupling_map: list,
            derivative_gates: list[str] = None,
            status: BackendStatus = None,
            supported_instructions: list[str] = None,
            credits_required: bool = None,
            online_date: datetime = None,
            display_name: str = None,
            description: str = None,
            tags: list = None,
            backend_type=None,
            machine_config=None,
            **kwargs,
    ):
        """Initializes the BackendConfiguration instance."""
        self.backend_id = backend_id
        self.backend_name = backend_name
        self.n_qubits = n_qubits
        self.basis_gates = basis_gates
        self.derivative_gates = derivative_gates
        self.gates = gates
        self.simulator = simulator
        self.conditional = conditional
        self.coupling_map = coupling_map
        self.local = local
        self.status = status
        self.supported_instructions = supported_instructions
        self.credits_required = credits_required
        self.online_date = online_date
        self.display_name = display_name
        self.description = description
        self.tags = tags
        self.properties = kwargs
        self.machine_config = machine_config
        self.backend_type = backend_type

    @staticmethod
    def from_dict(data: dict[str: str | int | list]) -> "BackendConfiguration":
        """Creates a BackendConfiguration instance from a dictionary.

        Args:
            data (dict): The dictionary containing backend configuration data.

        Returns:
            BackendConfiguration: The created BackendConfiguration instance.
        """
        return BackendConfiguration(**data)

    @staticmethod
    def from_api(data: dict[str: str | int | list]) -> "BackendConfiguration":
        """Creates a BackendConfiguration instance from API data.

        Args:
            data (dict): The API data containing backend configuration.

        Returns:
            BackendConfiguration: The created BackendConfiguration instance.
        """
        disabled_qubits = [q for q in data['disabledQubits'].split(',') if q]
        disabled_couplers = [g for g in data['disabledCouplers'].split(',') if g]
        backend_id = data['id']
        n_qubits = data['bitWidth']

        if data['labels'] == '1':
            backend_type = BackendType.quantum_computer
        else:
            backend_type = BackendType.simulator

        if backend_type == BackendType.quantum_computer:
            qpu = ApiClient().get_quantum_computer_config(backend_id)
            qubits = [int(q[1:]) for q in qpu['qubits'] if q not in disabled_qubits]
            coupling_map = []

            for k, q in qpu['coupler_map'].items():
                if k in disabled_couplers:
                    continue
                q0, q1 = q
                if q0 in disabled_qubits or q1 in disabled_qubits:
                    continue
                coupling_map.append([int(q0[1:]), int(q1[1:])])
        else:
            qubits = [i for i in range(n_qubits)]
            # todo: 全振幅仿真机，比特太多了。先限制一下
            coupling_map = [[i, j] for i in range(min(n_qubits, 100)) for j in range(i)]
        basis_gates = []
        derivative_gates = []
        gates = []

        for gate in data['baseGate']:
            name = gate['qcis'].lower()
            rule = gate['rule']
            # to qiskit gate name
            if name == 'i':
                name = 'id'
            elif name == 'b':
                name = 'barrier'
            elif name == 'm':
                name = 'measure'
            basis_gates.append(name)

            try:
                rule = json.loads(rule)
            except json.JSONDecodeError:
                pass

            # coupler gate
            if isinstance(rule, dict) and 'topology' in rule:
                gate_coupling_map = coupling_map
            # single gate
            else:
                gate_coupling_map = [[q] for q in qubits]

            gates.append(GateConfig(
                name=name,
                params=[f'p_{i}' for i in range(gate_parameters.get(name, 0))],
                coupling_map=gate_coupling_map
            ))

        for gate in data['derivativeGate']:
            name = gate['qcis'].lower()
            if name not in basis_gates:
                derivative_gates.append(name)

        data = {
            'backend_id': backend_id,
            'backend_name': data['code'],
            'n_qubits': data['bitWidth'],
            'basis_gates': basis_gates,
            'derivative_gates': derivative_gates,
            'gates': gates,
            'local': False,
            'simulator': backend_type in [BackendType.simulator, BackendType.offline_simulator],
            'supported_instructions': basis_gates,
            'credits_required': data['isToll'] == 2,
            'status': BackendStatus(data['status']),
            'online_date': datetime.strptime(data['createTime'], '%Y-%m-%d %H:%M:%S'),
            'display_name': data['name'],
            'description': '',
            'conditional': False,
            'coupling_map': coupling_map,
            'backend_type': backend_type,
        }
        return BackendConfiguration.from_dict(data)


class TianYanBackend(Backend):
    """Base class for TianYan backends."""

    def __init__(
            self,
            configuration: BackendConfiguration,
            api_client: 'ApiClient',
    ) -> None:
        """Initializes the TianYanBackend instance.

        Args:
            configuration (BackendConfiguration): The configuration of the backend.
            api_client (ApiClient): The client for interacting with the API.
        """
        super().__init__(name=configuration.backend_name, )
        self.configuration = configuration
        self.resource_id = configuration.backend_id
        self.resource_type = configuration.backend_id
        self.simulator = configuration.simulator
        self._api_client = api_client
        self._target = None
        self._machine_config = {}

    @classmethod
    def _default_options(cls):
        """Returns the default options for the backend.

        Returns:
            Options: The default options.
        """
        return Options()

    @property
    def max_circuits(self):
        """Returns the maximum number of circuits that can be executed in a single job.

        Returns:
            int: The maximum number of circuits.
        """
        return 50

    @property
    def machine_config(self):
        """Returns the machine configuration.

        Returns:
            dict: The machine configuration.
        """
        return self._machine_config

    @property
    def backend_type(self):
        """Returns the type of the backend.

        Returns:
            BackendType: The type of the backend.
        """
        return self.configuration.backend_type

    def run(
            self,
            run_input,
            shots: int = 1024,
            readout_calibration: bool = True,
            **options
    ) -> JobV1:
        """Submits a job to the backend.

        Args:
            run_input (QuantumCircuit | list[QuantumCircuit]): The circuit(s) to execute.
            shots (int, optional): The number of shots to execute. Defaults to 1024.
            readout_calibration (bool, optional): Whether to perform readout calibration. Defaults to True.
            **options: Additional options for the job.

        Returns:
            JobV1: The submitted job.

        Raises:
            TypeError: If the input type is not supported.
        """
        if isinstance(run_input, QuantumCircuit):
            circuits = [run_input]
        elif isinstance(run_input, list):
            circuits = run_input
        else:
            raise TypeError(f"Unsupported input type: {type(run_input)}")

        task_ids = self._api_client.submit_job(
            [to_cqlib(circ) for circ in circuits],
            machine=self.configuration.backend_name,
            shots=shots,
        )
        return TianYanJob(
            backend=self,
            job_id=','.join(task_ids),
            api_client=self._api_client,
            shots=shots,
            readout_calibration=readout_calibration,
            **options
        )

    @property
    def target(self):
        """Returns the target for the backend.

        Returns:
            Target: The target for the backend.
        """
        return self._target

    def __repr__(self) -> str:
        """Returns a string representation of the backend.

        Returns:
            str: The string representation.
        """
        return f"<{self.__class__.__name__}('{self.name}')>"


time_units = {
    's': 1,
    'ms': 1e-3,
    'us': 1e-6,
    'ns': 1e-9
}
frequency_units = {
    'hz': 1,
    'khz': 1e3,
    'mhz': 1e6,
    'ghz': 1e9
}
number_units = {
    '%': 1e-2,
    '': 1
}


class TianYanQuantumBackend(TianYanBackend):
    """Class representing a quantum computer backend on the TianYan platform."""

    def __init__(
            self,
            configuration: BackendConfiguration,
            api_client: 'ApiClient',
    ) -> None:
        """Initializes the TianYanQuantumBackend instance.

        Args:
            configuration (BackendConfiguration): The configuration of the backend.
            api_client (ApiClient): The client for interacting with the API.
        """
        super().__init__(configuration=configuration, api_client=api_client)
        self._machine_config = self._api_client.get_quantum_machine_config(
            self.configuration.backend_name
        )
        target = Target(
            # num_qubits=configuration.n_qubits,
            description=configuration.backend_name,
            qubit_properties=self._make_qubit_properties()
        )
        self._update_cz_gate(target)
        self._update_single_gates(target)
        self._update_measure_gate(target)
        self._update_barrier_gate(target)
        self._target = target

    def _make_qubit_properties(self):
        """Creates qubit properties from the machine configuration.

        Returns:
            list[QubitProperties | None]: The list of qubit properties.
        """
        t1 = self._machine_config['qubit']['relatime']['T1']
        t1_qubits = t1['qubit_used']
        t1_values = t1['param_list']
        t1_unit = time_units.get(t1['unit'].lower())
        t2 = self._machine_config['qubit']['relatime']['T2']
        t2_qubits = t2['qubit_used']
        t2_values = t2['param_list']
        t2_unit = time_units.get(t2['unit'].lower())
        frequency = self._machine_config['qubit']['frequency']['f01']
        frequency_qubits = frequency['qubit_used']
        frequency_values = frequency['param_list']
        frequency_unit = frequency_units.get(frequency['unit'].lower())

        if t1_qubits != t2_qubits != frequency_qubits:
            raise ValueError("t1/t2/frequency qubits are not the same")
        qubit_properties: list[QubitProperties | None] = [
            None for _ in range(self.configuration.n_qubits)
        ]
        for i, q in enumerate(t1_qubits):
            qubit_properties[int(q[1:])] = QubitProperties(
                t1=t1_values[i] * t1_unit,
                t2=t2_values[i] * t2_unit,
                frequency=frequency_values[i] * frequency_unit
            )
        return qubit_properties

    def _update_cz_gate(self, target: Target):
        """Updates the CZ gate in the target.

        Args:
            target (Target): The target to update.
        """
        cz_props = {}
        coupler_map = self._machine_config['overview']['coupler_map']
        gate_errors = self._machine_config['twoQubitGate']['czGate']['gate error']
        error_qubits = gate_errors['qubit_used']
        error_values = gate_errors['param_list']
        error_unit = number_units[gate_errors['unit']]

        for i, q in enumerate(error_qubits):
            q0, q1 = coupler_map[q]
            q0, q1 = int(q0[1:]), int(q1[1:])
            p = InstructionProperties(error=error_values[i] * error_unit, duration=1e-8)
            cz_props[q0, q1] = p
            cz_props[q1, q0] = p
        if 'cz' in self.configuration.basis_gates:
            target.add_instruction(CZGate(), cz_props)

    def _update_single_gates(self, target: Target):
        """Updates the single-qubit gates in the target.

        This method adds single-qubit gates (e.g., RZ, X2P, X2M, Y2P, Y2M, XY2P, XY2M)
        to the target based on the backend's configuration. It also includes the HGate
        and GlobalPhaseGate.

        Args:
            target (Target): The target to update with single-qubit gates.
        """
        rz_props = {}
        single_props = {}

        gate_params = self._machine_config['qubit']['singleQubit']['gate error']
        gate_values = gate_params['param_list']
        gate_qubits = gate_params['qubit_used']
        gate_unit = number_units[gate_params['unit']]

        for i, q in enumerate(gate_qubits):
            q_index = (int(q[1:]),)
            rz_props[q_index] = InstructionProperties(error=0, duration=0)
            single_props[q_index] = InstructionProperties(
                error=gate_values[i] * gate_unit,
                duration=0
            )
        if 'rz' in self.configuration.basis_gates:
            target.add_instruction(RZGate(Parameter('theta')), rz_props)
        if 'x2p' in self.configuration.basis_gates:
            target.add_instruction(X2PGate(), single_props.copy())
            # HGate is very import.
            target.add_instruction(HGate(), single_props.copy())
        if 'x2m' in self.configuration.basis_gates:
            target.add_instruction(X2MGate(), single_props.copy())
        if 'y2p' in self.configuration.basis_gates:
            target.add_instruction(Y2PGate(), single_props.copy())
        if 'y2m' in self.configuration.basis_gates:
            target.add_instruction(Y2MGate(), single_props.copy())
        if 'xy2p' in self.configuration.basis_gates:
            target.add_instruction(XY2PGate(Parameter('theta')), single_props.copy())
        if 'xy2m' in self.configuration.basis_gates:
            target.add_instruction(XY2MGate(Parameter('theta')), single_props.copy())

        target.add_instruction(GlobalPhaseGate(Parameter('phase')), {(): None})

    def _update_measure_gate(self, target: Target):
        """Updates the measurement gate in the target.

        This method adds the measurement gate to the target based on the backend's
        configuration and readout error data.

        Args:
            target (Target): The target to update with the measurement gate.
        """
        gate_params = self._machine_config['readout']['readoutArray']['Readout Error']
        gate_values = gate_params['param_list']
        gate_qubits = gate_params['qubit_used']
        gate_unit = number_units[gate_params['unit']]
        props = {
            (int(q[1:]),): InstructionProperties(error=gate_values[i] * gate_unit, duration=0)
            for i, q in enumerate(gate_qubits)
        }
        target.add_instruction(Measure(), props)

    def _update_barrier_gate(self, target: Target):
        """Updates the barrier gate in the target.

        This method adds the barrier gate to the target if it is supported by the
        backend's configuration.

        Args:
            target (Target): The target to update with the barrier gate.
        """
        if 'barrier' in self.configuration.basis_gates:
            target.add_instruction(Barrier, name="barrier")


class TianYanSimulatorBackend(TianYanBackend):
    """Class representing a simulator backend on the TianYan platform.

    This class extends the TianYanBackend to handle simulator-specific configurations
    and operations.
    """

    def __init__(
            self,
            configuration: BackendConfiguration,
            api_client: 'ApiClient',
    ) -> None:
        """Initializes the TianYanSimulatorBackend instance.

        Args:
            configuration (BackendConfiguration): The configuration of the simulator backend.
            api_client (ApiClient): The client for interacting with the API.
        """
        super().__init__(configuration=configuration, api_client=api_client)
        target = Target(
            num_qubits=configuration.n_qubits,
            description=configuration.backend_name,
        )
        self._update_gates(target)
        self._target = target

    def _update_gates(self, target):
        """Updates the gates in the target for the simulator backend.

        This method adds all supported gates (single-qubit, two-qubit, and measurement gates)
        to the target based on the backend's configuration.

        Args:
            target (Target): The target to update with the gates.
        """
        q_props = {(q,): None for q in range(self.configuration.n_qubits)}
        gates = set(self.configuration.basis_gates + self.configuration.derivative_gates)
        ins_mapping_list = {
            'rx': [standard_gates.RXGate(Parameter('theta')), q_props],
            'ry': [standard_gates.RYGate(Parameter('theta')), q_props],
            'rz': [standard_gates.RZGate(Parameter('theta')), q_props],
            'x2p': [X2PGate(), q_props],
            'x2m': [X2MGate(), q_props],
            'y2p': [Y2PGate(), q_props],
            'y2m': [Y2MGate(), q_props],
            'xy2p': [XY2PGate(Parameter('theta')), q_props],
            'xy2m': [XY2MGate(Parameter('theta')), q_props],
            'h': [standard_gates.HGate(), q_props],
            'x': [standard_gates.XGate(), q_props],
            'y': [standard_gates.YGate(), q_props],
            'z': [standard_gates.ZGate(), q_props],
            's': [standard_gates.SGate(), q_props],
            'sd': [standard_gates.SdgGate(), q_props],
            't': [standard_gates.TGate(), q_props],
            'td': [standard_gates.TdgGate(), q_props],

            'measure': [Measure(), q_props],
        }
        ins_mapping_dict = {
            'cz': {'instruction': CZGate(), 'properties': {None: None}},
            'cx': {'instruction': CXGate(), 'properties': {None: None}},
            'barrier': {'instruction': Barrier, 'name': 'barrier'}
        }
        for gate in gates:
            if gate in ins_mapping_list:
                target.add_instruction(*ins_mapping_list[gate])
            elif gate in ins_mapping_dict:
                target.add_instruction(**ins_mapping_dict[gate])
            else:
                warnings.warn(f'{gate} is not supported in simulator backend.')



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
API Client module for interacting with quantum computing platforms.
"""
import os
import time
from datetime import datetime

import requests
from cqlib.circuits import Circuit
from cqlib.quantum_platform import TianYanPlatform, QuantumLanguage
from cqlib.exceptions import CqlibRequestError


class ApiClient:
    """
    A client for managing API interactions with a quantum computing platform.

    Attributes
    ----------
    base_url : str
        The base URL for the API endpoints.
    backends : str
        The endpoint for fetching available quantum backends.
    quantum_computer_config : str
        The endpoint for fetching all quantum computer configurations.
    quantum_machine_config : str
        The endpoint for fetching quantum machine configurations.
    query_experiment_result : str
        The endpoint for querying experiment results.
    """
    base_url = 'https://qc.zdxlz.com'
    backends = "/qccp-quantum/experiments/quantum/computer/by/user"
    quantum_computer_config = "/qccp-quantum/experiments/quantum/computer/config"
    quantum_machine_config = '/qccp-quantum/sdk/experiment/download/config'
    query_experiment_result = '/qccp-quantum/sdk/experiment/result/find'

    def __init__(self, token: str = None, base_url: str | None = None) -> None:
        """
        Initializes the ApiClient with an optional token and base URL.

        Args:
            token(str): The authentication token for the API. If not provided,
                it will be fetched from the environment variable CQLIB_TOKEN.
            base_url(str): The base URL for the API endpoints. If not provided,
                the default URL will be used.
        """
        if base_url is not None:
            self.base_url = base_url
        if token is None:
            token = os.environ.get("CQLIB_TOKEN", "")
        self.session = requests.Session()
        self._token = token
        if token is not None:
            self._pf = TianYanPlatform(token)
        else:
            self._pf = None

    @property
    def platform(self) -> TianYanPlatform:
        """
        Returns the TianYanPlatform instance associated with the client.

        Returns:
            TianYanPlatform: The platform instance for managing quantum
                computing tasks.
        """
        if self._pf is None:
            self._pf = TianYanPlatform(self._token)
        return self._pf

    def get_backends(self) -> list[dict]:
        """
        Fetches the list of available quantum backends.

        Returns:
            list[dict]: A list of dictionaries containing information about
                available quantum backends.
        """
        response = self.session.get(
            f'{self.base_url}{self.backends}',
            params={'a': str(int(time.time())), 'apiCode': 'byUser', },
            headers={
                'Accept': 'application/json, text/plain, */*',
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:137.0) Gecko/20100101 Firefox/137.0',
                'apiCode': 'byUser',
                'requestTime': self._request_time()
            })
        data = response.json()
        if data.get('code') != 200:
            raise CqlibRequestError(f"request error. \n{data.get('message')}")
        return data.get('data', [])

    def get_quantum_computer_config(self, computer_id) -> dict:
        """
        Fetches the configuration of a specific quantum computer.

        Args:
            computer_id(str): The ID of the quantum computer.

        Returns:
            dict: A dictionary containing the configuration details of the quantum computer.
        """
        response = self.session.get(
            f'{self.base_url}{self.quantum_computer_config}',
            params={
                'a': str(int(time.time())),
                'type': 'overview',
                'quantumComputerId': computer_id,
                'label': 'qubits,couplers,coupler_map,disabled_qubits,disabled_couplers'
            },
            headers={
                'Accept': 'application/json, text/plain, */*',
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:137.0) Gecko/20100101 Firefox/137.0',
                'apiCode': 'config',
                'requestTime': self._request_time()
            })
        data = response.json()
        return data.get('data', {})

    def get_quantum_machine_config(self, computer_code: str) -> dict:
        """
        Downloads the physical machine parameters for a quantum machine (not a simulator).

        Args:
            computer_code(str): The code of the quantum machine.

        Returns:
            dict: A dictionary containing the configuration details of the quantum machine.
        """
        return self.platform.download_config(machine=computer_code)

    def submit_job(
            self,
            circuits: Circuit | list[Circuit],
            machine: str,
            **kwargs
    ) -> list[str]:
        """
        Submits a job (or list of jobs) to a specified quantum machine.

        Args:
            circuits(Circuit | list[Circuit]): The quantum circuit(s) to be executed.
            machine(str): The name of the quantum machine to execute the job.
            **kwargs: Additional keyword arguments, such as the number of shots.

        Returns:
            list[str]: A list of task IDs for the submitted jobs.
        """
        if not isinstance(circuits, list):
            circuits = [circuits]

        task_ids = self.platform.submit_experiment(
            [c.as_str() for c in circuits],
            machine_name=machine,
            language=QuantumLanguage.QCIS,
            num_shots=kwargs.get('shots', 1000),
        )
        return task_ids

    def query_job(self, task_ids: list[str]) -> list[dict]:
        """
        Queries the results of submitted jobs using their task IDs.

        Args:
            task_ids(list[str]): A list of task IDs to query.

        Returns:
            list[dict]: A list of dictionaries containing the results of the queried jobs.

        Raises:
        CqlibRequestError: If the request to query job results fails.
        """
        response = self.session.post(
            f'{self.base_url}{self.query_experiment_result}',
            json={"query_ids": task_ids},
            headers={
                'Accept': 'application/json, text/plain, */*',
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:137.0) Gecko/20100101 Firefox/137.0',
                'apiCode': 'config',
                'requestTime': self._request_time(),
                "basicToken": self.platform.access_token,
                "Authorization": f'Bearer {self.platform.access_token}'
            })
        data = response.json()
        if data.get('code') != 0:
            raise CqlibRequestError(f"request error {data.get('msg')}")
        return data.get('data').get('experimentResultModelList')

    @staticmethod
    def _request_time() -> str:
        """
        Generates a formatted timestamp for API requests.
        """
        return datetime.now().strftime("%m/%d/%Y, %I:%M:%S %p") \
            .replace("/0", "/").lstrip("0").replace(" 0", " ")



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Sampler module for executing quantum circuits on the TianYan backend.

This module provides a class to sample quantum circuits using the TianYan backend,
leveraging Qiskit's primitives for efficient execution.
"""

from qiskit.primitives import BackendSamplerV2
from qiskit.providers import BackendV2 as Backend, Options

from .tianyan_backend import TianYanBackend


class TianYanSampler(BackendSamplerV2):
    """A sampler class for executing quantum circuits on the TianYan backend.

    This class extends Qiskit's `BackendSamplerV2` to provide sampling functionality
    specifically for the TianYan backend.
    """

    def __init__(self, backend: TianYanBackend, options: dict | None = None) -> None:
        """Initializes the TianYanSampler instance.

         Args:
             backend (TianYanBackend): The TianYan backend to use for sampling.
             options (dict, optional): Additional options for the sampler. Defaults to None.
         """
        super().__init__(backend=backend, options=options)

    @property
    def backend(self) -> Backend:
        """Returns the backend associated with the sampler.

        Returns:
            Backend: The backend used for executing quantum circuits.
        """
        return self._backend



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
TianYan Provider module for managing quantum computing backends.

This module provides a class to interact with the TianYan platform, allowing users to
retrieve and manage quantum computing backends (both quantum computers and simulators).
"""

import os
from pathlib import Path

import dotenv

from .tianyan_backend import TianYanBackend, BackendConfiguration, CqlibAdapterError, \
    BackendStatus, TianYanQuantumBackend, TianYanSimulatorBackend
from .api_client import ApiClient


class TianYanProvider:
    """A provider class for interacting with the TianYan quantum computing platform.

    This class allows users to retrieve and manage backends (quantum computers and simulators)
    available on the TianYan platform.
    """

    def __init__(
            self,
            token: str | None = None,
            *,
            load_dotenv: bool = True,
            dotenv_path: str | Path | None = None,
    ) -> None:
        """Initializes the TianYanProvider instance.

        Args:
            token (str, optional): The authentication token for the TianYan platform.
                If not provided, it will be fetched from the environment variable CQLIB_TOKEN.
            load_dotenv (bool, optional): Whether to load environment variables from a .env file.
                Defaults to True.
            dotenv_path (str | Path, optional): The path to the .env file. Defaults to None.
        """
        if load_dotenv or dotenv_path is not None:
            dotenv.load_dotenv(dotenv_path)

        if token is None:
            token = os.environ.get("CQLIB_TOKEN", "")
        self.token = token

        self.name = "qiskit_adapter"
        self._api_client = ApiClient(token=token)

    def backends(self, simulator: bool = None, online: bool = True, name: str = None):
        """Retrieves a list of backends based on the specified filters.

        Args:
            simulator (bool, optional): Whether to filter for simulator backends.
                If None, no filtering is applied. Defaults to None.
            online (bool, optional): Whether to filter for online backends.
                Defaults to True.
            name (str, optional): The name of the backend to filter for.
                If None, no filtering is applied. Defaults to None.

        Returns:
            list[TianYanBackend]: A list of backends matching the specified filters.
        """
        bs = []
        for data in self._api_client.get_backends():
            cfg = BackendConfiguration.from_api(data)
            if online and cfg.status not in [BackendStatus.running, BackendStatus.calibrating]:
                continue
            if simulator is not None and cfg.simulator != simulator:
                continue
            if name is not None and cfg.backend_name != name:
                continue
            if cfg.simulator:
                backend = TianYanSimulatorBackend(
                    configuration=cfg,
                    api_client=self._api_client,
                )
            else:
                backend = TianYanQuantumBackend(
                    configuration=cfg,
                    api_client=self._api_client,
                )
            bs.append(backend)
        return bs

    def backend(self, name: str) -> TianYanBackend:
        """Retrieves a specific backend by name.

        Args:
            name (str): The name of the backend to retrieve.

        Returns:
            TianYanBackend: The backend with the specified name.

        Raises:
            CqlibAdapterError: If the backend with the specified name is not found.
        """
        for b in self.backends():
            if b.name == name:
                return b
        raise CqlibAdapterError(f"Backend {name} not found")



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Job module for managing quantum computing jobs.

This module provides a class to handle job submission, status checking, and result retrieval
for quantum computing tasks executed on a backend.
"""

import time
from collections import defaultdict

import numpy as np
from qiskit.providers import JobV1, JobStatus
from qiskit.result import Result

from qiskit.providers.backend import Backend

from cqlib.utils.laboratory_utils import LaboratoryUtils
from .api_client import ApiClient


class TianYanJob(JobV1):
    """A class representing a job executed on the TianYan quantum computing platform.

    Attributes:
        job_id (str): The unique identifier for the job.
        backend (Backend): The backend where the job is executed.
        api_client (ApiClient): The client for interacting with the API.
    """
    def __init__(
            self,
            job_id: str,
            backend: Backend | None = None,
            api_client: ApiClient | None = None,
            **kwargs
    ) -> None:
        """Initializes the TianYanJob instance.

        Args:
            job_id (str): The unique identifier for the job.
            backend (Backend, optional): The backend where the job is executed. Defaults to None.
            api_client (ApiClient, optional): The client for interacting with the API. Defaults to None.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(backend=backend, job_id=job_id, **kwargs)
        if api_client is None:
            api_client = ApiClient()
        self._api_client = api_client

    def submit(self):
        """Submit the job to the backend for execution."""
        pass

    def result(self) -> Result:
        """Retrieves the results of the job.

        Returns:
            Result: The results of the job, including counts and memory data.

        Raises:
            Exception: If the job fails to complete or retrieve results.
        """
        task_ids = self._job_id.split(',')
        while 1:
            status = self.status()
            if status == JobStatus.DONE:
                break
            time.sleep(1)

        lu = LaboratoryUtils()
        results = []
        for item in self._api_client.query_job(task_ids):
            readout_calibration = self.metadata.get('readout_calibration', True)
            if 'shots' in self.metadata:
                shots = self.metadata['shots']
            else:
                shots = len(item['resultStatus']) - 1
            if self._backend.simulator or not readout_calibration:
                basis_list = lu.readout_data_to_state_probabilities(item)
                counts, memory_list = self.to_counts(basis_list)
            else:
                calibration_result = lu.probability_calibration(
                    result=item, laboratory=None, config_json=self._backend.machine_config
                )
                counts, memory_list = self.probability_correction(calibration_result, shots)

            results.append({
                'shots': shots,
                'success': True,
                'data': {
                    'counts': counts,
                    'memory': memory_list
                }
            })

        return Result.from_dict({
            "backend_name": self._backend.configuration.backend_name,
            "backend_version": self._backend.version,
            "qobj_id": id(self._job_id),
            "job_id": self._job_id,
            "success": True,
            'status': JobStatus.DONE,
            "results": results,
        })

    def cancel(self):
        """Attempts to cancel the job.

        Raises:
            NotImplementedError: If cancellation is not supported.
        """
        raise NotImplementedError

    def status(self) -> JobStatus:
        """Retrieves the status of the job.

        Returns:
            JobStatus: The current status of the job (e.g., DONE, RUNNING, QUEUED).
        """
        task_ids = self._job_id.split(',')
        # todo: use new light api
        data = self._api_client.query_job(task_ids)
        ld = len(data)

        if ld == len(task_ids):
            return JobStatus.DONE
        if ld > 0:
            return JobStatus.RUNNING
        return JobStatus.QUEUED

    @staticmethod
    def to_counts(state01):
        """Calculates the probability distribution of measurement outcomes for a given quantum state.

        Args:
            state01 (list): A one-dimensional or two-dimensional list representing the quantum state, with complex numbers as elements.

        Returns:
            dict: A dictionary where keys are binary strings representing measurement outcomes and values are their corresponding probabilities.
        """
        if isinstance(state01[0], bool):
            state01 = [state01]
        counts = defaultdict(int)
        state_01_t = np.transpose(state01)
        memory_list = []
        for num in range(len(state_01_t)):
            k = 0
            for i in range(len(state01)):
                k += state_01_t[num][i] * (2 ** i)
            prob_state = hex(k)
            counts[prob_state] += 1
            memory_list.append(prob_state)

        return counts, memory_list

    @staticmethod
    def probability_correction(probabilities, shots):
        """Corrects the measured probability of quantum states.

        Args:
            probabilities (dict): The measured probabilities of quantum states.
            shots (int): The number of shots (measurements) taken.

        Returns:
            dict: A dictionary of corrected probabilities.
            list: A list of memory states after correction.
        """
        hex_keys = []
        norm_probs = []
        for k, v in probabilities.items():
            hex_keys.append(hex(int(k, 2)))
            norm_probs.append(v)

        probs = np.clip(norm_probs, 0, 1)
        probs /= probs.sum()

        exact_counts = (shots * probs).astype(np.float64)
        integer_part = exact_counts.astype(int)
        fractional_part = exact_counts - integer_part
        remaining = shots - integer_part.sum()
        if remaining > 0:
            # 按小数部分大小分配剩余计数
            indices = np.argpartition(-fractional_part, remaining)[:remaining]
            integer_part[indices] += 1
        counts_dict = {k: c for k, c in zip(hex_keys, integer_part) if c > 0}
        memory_list = np.repeat(hex_keys, integer_part)
        np.random.shuffle(memory_list)
        return counts_dict, memory_list.tolist()



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Version number"""
import os
import sys

if sys.version_info < (3, 10, 0):  # pragma: no cover
    raise SystemError(
        "cqlib requires Python 3.10 or higher. \n"
        "Please upgrade your Python version to continue using this library. \n"
        "You can download the latest version of Python at https://www.python.org/downloads/\n"
    )

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(ROOT_DIR, "VERSION.txt")) as version_file:
    VERSION = version_file.read().strip()

version: str
__version__: str

__version__ = version = VERSION



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Gates module for defining custom quantum gates and their equivalence rules.

This module provides custom quantum gates (e.g., X2P, X2M, Y2P, Y2M, XY2P, XY2M)
and their equivalence rules with standard Qiskit gates. It also includes
validation to ensure the correctness of the gate definitions.
"""

from math import pi

import numpy as np
from qiskit.quantum_info import Operator
from qiskit.circuit import Gate, QuantumCircuit, Parameter
from qiskit.circuit.library import GlobalPhaseGate, RXGate, RYGate, \
    HGate, XGate, YGate
from qiskit.circuit.equivalence_library import SessionEquivalenceLibrary as SELib

from cqlib.circuits.gates.x import X2P, X2M
from cqlib.circuits.gates.y import Y2P, Y2M
from cqlib.circuits.gates.xy import XY2P, XY2M
from cqlib.circuits.gates.rxy import RXY


class X2PGate(Gate):
    """
    Custom quantum gate representing a positive X rotation by π/2.

    This gate is equivalent to a rotation around the X-axis by π/2 radians.
    """

    def __init__(self, label=None):
        """
        Initializes the X2PGate.

        Args:
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("x2p", 1, params=[], label=label)

    def _define(self):
        """Defines the quantum circuit for the X2P gate."""
        defn = QuantumCircuit(1)
        defn.rx(pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(X2P(), dtype=dtype)


class X2MGate(Gate):
    """
    Custom quantum gate representing a negative X rotation by π/2.

    This gate is equivalent to a rotation around the X-axis by -π/2 radians.
    """

    def __init__(self, label=None):
        """
        Initializes the X2MGate.

        Args:
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("x2m", 1, [], label=label)

    def _define(self):
        """Defines the quantum circuit for the X2M gate."""
        defn = QuantumCircuit(1)
        defn.rx(-pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(X2M(), dtype=dtype)


class Y2PGate(Gate):
    """
    Custom quantum gate representing a positive Y rotation by π/2.

    This gate is equivalent to a rotation around the Y-axis by π/2 radians.
    """

    def __init__(self, label=None):
        """
        Initializes the Y2PGate.

        Args:
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("y2p", 1, params=[], label=label)

    def _define(self):
        """Defines the quantum circuit for the Y2P gate."""
        defn = QuantumCircuit(1)
        defn.ry(pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(Y2P(), dtype=dtype)


class Y2MGate(Gate):
    """
    Custom quantum gate representing a negative Y rotation by π/2.

    This gate is equivalent to a rotation around the Y-axis by -π/2 radians.
    """

    def __init__(self, label=None):
        """
        Initializes the Y2MGate.

        Args:
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("y2m", 1, [], label=label)

    def _define(self):
        """Defines the quantum circuit for the Y2M gate."""
        defn = QuantumCircuit(1)
        defn.ry(-pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(Y2M(), dtype=dtype)


class XY2PGate(Gate):
    """
    Custom quantum gate representing a positive XY rotation.

    This gate is parameterized by an angle `theta` and represents a rotation
    in the XY plane.
    """

    def __init__(self, theta: float | Parameter, label: str = None):
        """
        Initializes the XY2PGate.

        Args:
            theta (float|Parameter): The rotation angle.
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("xy2p", 1, [theta], label=label)

    def _define(self):
        """Defines the quantum circuit for the XY2P gate."""
        theta_ = self.params[0]
        defn = QuantumCircuit(1)
        defn.rz(pi / 2 - theta_, 0)
        defn.ry(pi / 2, 0)
        defn.rz(theta_ - pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(XY2P(self.params[0]), dtype=dtype)


class XY2MGate(Gate):
    """
    Custom quantum gate representing a negative XY rotation.

    This gate is parameterized by an angle `theta` and represents a rotation
    in the XY plane.
    """

    def __init__(self, theta: float | Parameter, label: str = None):
        """
        Initializes the XY2MGate.

        Args:
            theta (float | Parameter): The rotation angle.
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("xy2m", 1, [theta], label=label)

    def _define(self):
        """Defines the quantum circuit for the XY2M gate."""
        theta_ = self.params[0]
        defn = QuantumCircuit(1)
        defn.rz(-pi / 2 - theta_, 0)
        defn.ry(pi / 2, 0)
        defn.rz(theta_ + pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(XY2M(self.params[0]), dtype=dtype)


class RxyGate(Gate):
    """A parameterized two-axis rotation gate in the XY plane.

    This gate implements a composite rotation consisting of Rz and Rx operations.
    The rotation is parameterized by two angles (phi and theta) which control
    the Z-axis rotations before and after the X-axis π/2 rotations.

    Args:
        phi (float | Parameter): Rotation angle (in radians) for initial Z-axis rotation.
            Controls the phase offset in the composite rotation sequence.
        theta (float | Parameter): Rotation angle (in radians) for middle Z-axis rotation.
            Determines the main rotation magnitude between X-axis operations.
        label (str, optional): Optional text label for gate identification. Defaults to None.

    Example:
        >>> import math
       >>> from qiskit import QuantumCircuit
        >>> qc = QuantumCircuit(1)
        >>> qc.append(RxyGate(math.pi/3, math.pi/4), [0])
    """

    def __init__(self, phi: float | Parameter, theta: float | Parameter, label: str = None):
        """
        Initializes the RxyGate.

        Args:
            phi (float | Parameter): The rotation angle.
            theta (float | Parameter): The rotation angle.
            label (str, optional): A custom label for the gate. Defaults to None.
        """
        super().__init__("rxy", 1, [phi, theta], label=label)

    def _define(self):
        """Defines the quantum circuit for the RXY gate.

        Implements gate using rotation sequence:
        Rz(π/2 - φ) → Rx(π/2) → Rz(θ) → Rx(-π/2) → Rz(φ - π/2)
        """
        phi_ = Parameter("phi")
        theta_ = Parameter("theta")
        defn = QuantumCircuit(1)
        defn.rz(pi / 2 - phi_, 0)
        defn.rx(pi / 2, 0)
        defn.rz(theta_, 0)
        defn.rx(-pi / 2, 0)
        defn.rz(phi_ - pi / 2, 0)
        self._definition = defn

    def __array__(self, dtype=None, copy=None):
        """
        Returns the matrix representation of the gate.

        Args:
            dtype: The data type of the array.
            copy: Whether to avoid copying the array.

        Returns:
            np.ndarray: The matrix representation of the gate.

        Raises:
            ValueError: If copying cannot be avoided.
        """
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        return np.asarray(RXY(self.params[0], self.params[1]), dtype=dtype)


x_qc = QuantumCircuit(1)
x_qc.rx(pi / 2, 0)
SELib.add_equivalence(X2PGate(), x_qc)

# x
x_qc = QuantumCircuit(1)
x_qc.append(X2PGate(), [0])
x_qc.append(X2PGate(), [0])
x_qc.append(GlobalPhaseGate(pi / 2), [])
SELib.add_equivalence(XGate(), x_qc)
assert np.allclose(np.asarray(XGate()), Operator(x_qc).to_matrix())

# RX
theta = Parameter("theta")
rx_qc = QuantumCircuit(1)
rx_qc.rz(pi / 2, 0)
rx_qc.append(X2PGate(), [0])
rx_qc.rz(theta, 0)
rx_qc.append(X2MGate(), [0])
rx_qc.rz(-pi / 2, 0)
SELib.add_equivalence(RXGate(theta), rx_qc)

# RY
theta = Parameter("theta")
ry_qc = QuantumCircuit(1)
ry_qc.append(X2PGate(), [0])
ry_qc.rz(theta, 0)
ry_qc.append(X2MGate(), [0])
SELib.add_equivalence(RYGate(theta), ry_qc)

# X2M
x2m_qc = QuantumCircuit(1)
x2m_qc.rx(-pi / 2, 0)
SELib.add_equivalence(X2MGate(), x2m_qc)

ry_qc = QuantumCircuit(1)
ry_qc.ry(pi / 2, 0)
SELib.add_equivalence(Y2PGate(), ry_qc)

# Y
y_qc = QuantumCircuit(1)
y_qc.append(Y2PGate(), [0])
y_qc.append(Y2PGate(), [0])
y_qc.append(GlobalPhaseGate(pi / 2), [])
SELib.add_equivalence(YGate(), y_qc)

ry_qc = QuantumCircuit(1)
ry_qc.ry(-pi / 2, 0)
SELib.add_equivalence(Y2MGate(), ry_qc)

#  H gate
h_decomp = QuantumCircuit(1)
h_decomp.rz(pi, 0)
h_decomp.append(Y2PGate(), [0])
h_decomp.append(GlobalPhaseGate(pi / 2), [])
SELib.add_equivalence(HGate(), h_decomp)

# XY2P
theta = Parameter("theta")
xy2p_qc = QuantumCircuit(1)
xy2p_qc.rz(pi / 2 - theta, 0)
xy2p_qc.ry(pi / 2, 0)
xy2p_qc.rz(theta - pi / 2, 0)
SELib.add_equivalence(XY2PGate(theta), xy2p_qc)

# XY2M
theta = Parameter("theta")
xy2m_qc = QuantumCircuit(1)
xy2m_qc.rz(-pi / 2 - theta, 0)
xy2m_qc.ry(pi / 2, 0)
xy2m_qc.rz(theta + pi / 2, 0)
SELib.add_equivalence(XY2MGate(theta), xy2m_qc)

phi = Parameter("phi")
theta = Parameter("theta")
qc = QuantumCircuit(1)
qc.rz(pi / 2 - phi, 0)
qc.rx(pi / 2, 0)
qc.rz(theta, 0)
qc.rx(-pi / 2, 0)
qc.rz(phi - pi / 2, 0)
SELib.add_equivalence(RxyGate(phi, theta), qc)

__all__ = [
    'X2PGate',
    'X2MGate',
    'Y2PGate',
    'Y2MGate',
    'XY2PGate',
    'XY2MGate',
    'RxyGate'
]



# This code is part of cqlib.
#
# Copyright (C) 2025 China Telecom Quantum Group.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Adapter module for integrating different interfaces.

This module provides classes and methods to adapt various interfaces,
making them compatible with the expected input/output format.
"""

from qiskit.circuit import QuantumCircuit as QiskitCircuit

from cqlib import Circuit as CqlibCircuit


def to_cqlib(qiskit_circuit: QiskitCircuit) -> CqlibCircuit:
    """
    Converts a Qiskit QuantumCircuit to a CqlibCircuit.

    Args:
        qiskit_circuit (QiskitCircuit): The Qiskit quantum circuit to be converted.

    Returns:
        CqlibCircuit: The converted circuit in the Cqlib format.

    Raises:
        NotImplementedError: If an operation in the Qiskit circuit is not supported
                             by the CqlibCircuit.
    """
    qubits_mapping = {q: i for i, q in enumerate(qiskit_circuit.qubits)}
    clbits_mapping = {c: i for i, c in enumerate(qiskit_circuit.clbits)}
    measure_qubits = {}
    cqlib_circuit = CqlibCircuit(list(qubits_mapping.values()))

    for ins in qiskit_circuit.data:
        operation = ins.operation
        qs = [qubits_mapping[q] for q in ins.qubits]
        cs = [clbits_mapping[c] for c in ins.clbits]

        if operation.name == "measure":
            measure_qubits[cs[0]] = qs[0]
        elif operation.name == "barrier":
            cqlib_circuit.barrier(*qs)
        elif operation.name == "i":
            cqlib_circuit.i(qs[0], 30)
        elif operation.name == 'unitary':
            ps = operation.params
            if len(ps) > 0:
                ps = ps[1:]
            getattr(cqlib_circuit, operation.label)(*qs, *ps)
        elif hasattr(cqlib_circuit, operation.name):
            getattr(cqlib_circuit, operation.name)(*qs, *operation.params)
        elif operation.name == 'tdg':
            cqlib_circuit.td(*qs, *operation.params)
        elif operation.name == 'sdg':
            cqlib_circuit.sd(*qs, *operation.params)
        else:
            raise NotImplementedError(f"{operation.name} is not supported")
    for cl_bit in sorted(measure_qubits.keys()):
        cqlib_circuit.measure(measure_qubits[cl_bit])

    return cqlib_circuit
